@@ -12,6 +12,7 @@ from typing import TYPE_CHECKING, cast
 
 from clu import Command, CommandStatus
 
+from hal import config
 from hal.exceptions import HALError
 
 
@@ -25,7 +26,6 @@ class HALHelper:
     name: str | None = None
 
     def __init__(self, actor: HALActor):
-
         self.actor = actor
 
     async def _send_command(
@@ -60,9 +60,20 @@ class HALHelper:
         return cast(Command, cmd)
 
 
+def get_default_exposure_time(observatory: str, design_mode: str | None = None):
+    """Returns the default exposure time for the current design mode."""
+
+    if design_mode is not None and "bright" in design_mode:
+        return config["macros"]["expose"]["fallback"]["exptime"]["bright_design_mode"][observatory.upper()]  # fmt: skip  # noqa
+    return config["macros"]["expose"]["fallback"]["exptime"]["default"]
+
+
 from .apogee import *
 from .boss import *
+from .cherno import *
 from .ffs import *
+from .jaeger import *
 from .lamps import *
+from .overhead import *
 from .scripts import *
 from .tcc import *
