@@ -23,7 +23,7 @@ from .macro import Macro
 # TODO: maybe Macro subclasses should be a singleton ...
 
 
-exclusions = ["__init__.py", "base.py"]
+exclusions = ["__init__.py", "macro.py"]
 
 cwd = os.getcwd()
 os.chdir(os.path.dirname(os.path.realpath(__file__)))
@@ -36,13 +36,15 @@ for f_ in files:
         modname = f_[0:-3].replace("/", ".")
         mod = importlib.import_module("hal.macros." + modname)
         for objname in dir(mod):
+            if objname.startswith("_"):
+                continue
+
             obj = getattr(mod, objname)
             if inspect.isclass(obj) and issubclass(obj, Macro) and obj != Macro:
                 all_macros.append(obj())
                 locals().update({objname: obj})
 
     except Exception as ee:
-        raise
         warnings.warn(f"cannot import file {f_}: {ee}", HALUserWarning)
 
 os.chdir(cwd)
