@@ -20,6 +20,7 @@ from clu import Command, CommandStatus
 
 from hal import config
 from hal.exceptions import HALUserWarning, MacroError
+from hal.helpers.overhead import OverheadHelper
 
 
 if TYPE_CHECKING:
@@ -32,6 +33,21 @@ __all__ = ["Macro"]
 StageType = Union[str, tuple[str, ...], list[str]]
 
 
+def record_overhead(macro: Macro):
+    """Runs a macro stage and records its overhead."""
+
+    async def record_overhead_wrapper(stage_coro: Coroutine[Any, Any, None]):
+        overhead_helper = OverheadHelper(
+            macro,
+            stage_coro.__name__,
+            macro_id=macro.macro_id,
+        )
+        async with overhead_helper:
+            await stage_coro
+
+    return record_overhead_wrapper
+
+
 class StageStatus(enum.Flag):
     """Stage status codes."""
 
@@ -56,25 +72,22 @@ def flatten(stages: list[StageType]) -> list[str]:
 class Macro:
     """A base macro class that offers concurrency and cancellation."""
 
-    __RUNNING__: ClassVar[list[str]] = []
+    name: ClassVar[str]
+    observatory: str | None = None
 
-    name: str
+    __RUNNING__: ClassVar[list[str]] = []
 
     __STAGES__: list[StageType]
     __PRECONDITIONS__: list[StageType] = []
     __CLEANUP__: list[StageType] = []
 
-    def __init__(self, name: Optional[str] = None):
-
-        if name is None and not hasattr(self, "name"):
-            raise MacroError("The macro does not have a name attribute.")
-        self.name = name or self.name
-
+    def __init__(self):
         if not hasattr(self, "__STAGES__"):
             raise MacroError("Must override __STAGES__.")
 
         self.stages = self.__PRECONDITIONS__ + self.__STAGES__ + self.__CLEANUP__
-        self._flat_stages = flatten(self.stages)
+
+        self.flat_stages = flatten(self.stages)
 
         for stage in self.__PRECONDITIONS__ + self.__CLEANUP__:
             if not isinstance(stage, str):
@@ -96,11 +109,15 @@ class Macro:
 
         self.command: HALCommandType
 
+        self.macro_id = OverheadHelper.get_next_macro_id()
+
         self._running: bool = False
         self.failed: bool = False
         self.cancelled: bool = False
 
         self._running_task: asyncio.Task | None = None
+        self._running_event = asyncio.Event()
+        self._running_event.set()  # Won't unset until the macro is actually running.
 
     def __repr__(self):
         stages = flatten(self.stages)
@@ -135,7 +152,9 @@ class Macro:
         if command is None:
             raise MacroError("A new command must be passed to reset.")
 
-        self._reset_internal(**opts)
+        self.command = command
+        if self.observatory is None and command.actor is not None:
+            self.observatory = command.actor.observatory
 
         if reset_stages is None:
             self.stages = self.__PRECONDITIONS__ + self.__STAGES__ + self.__CLEANUP__
@@ -171,7 +190,7 @@ class Macro:
         if len(self.stages) == 0:
             raise MacroError("No stages found.")
 
-        self._flat_stages = flatten(self.stages)
+        self.flat_stages = flatten(self.stages)
 
         # Reload the config and update it with custom options for this run.
         self.config = defaultdict(lambda: None, self._base_config.copy())
@@ -188,11 +207,15 @@ class Macro:
             if getattr(self, st, None) is None:
                 raise MacroError(f"Cannot find method for stage {st!r}.")
 
+        self._reset_internal(**opts)
+
         self.running = False
         self._running_task = None
 
-        if command:
-            self.command = command
+        self.macro_id = OverheadHelper.get_next_macro_id()
+
+        if not self._running_event.is_set():
+            self._running_event.set()
 
         self.list_stages()
 
@@ -229,6 +252,13 @@ class Macro:
 
         self.command.debug(running_macros=Macro.__RUNNING__)
 
+        if is_running:
+            if not self._running_event.is_set():
+                self._running_event.set()
+            self._running_event.clear()
+        else:
+            self._running_event.set()
+
     def set_stage_status(
         self,
         stages: StageType,
@@ -295,13 +325,16 @@ class Macro:
 
     async def fail_macro(
         self,
-        error: Exception,
+        error_or_message: Exception | str,
         stage: Optional[StageType] = None,
     ):
         """Fails the macros and informs the actor."""
 
-        self.command.error(error=error)
-        self.command.actor.log.exception(f"Macro {self.name} failed with error:")
+        if isinstance(error_or_message, Exception):
+            self.command.error(error=error_or_message)
+            self.command.actor.log.exception(f"Macro {self.name} failed with error:")
+        else:
+            self.command.info(text=error_or_message)
 
         self.failed = True
 
@@ -330,7 +363,12 @@ class Macro:
 
             try:
                 self.set_stage_status(cleanup_stage, StageStatus.ACTIVE)
-                await asyncio.gather(*self._get_coros(cleanup_stage))
+                await asyncio.gather(
+                    *[
+                        record_overhead(self)(coro)
+                        for coro in self._get_coros(cleanup_stage)
+                    ]
+                )
                 self.set_stage_status(cleanup_stage, StageStatus.FINISHED)
             except Exception as err:
                 self.command.error(f"Cleanup {cleanup_stage} failed: {err}")
@@ -339,7 +377,7 @@ class Macro:
 
         self.running = False
 
-    async def run(self):
+    async def run(self) -> bool:
         """Executes the macro allowing for cancellation."""
 
         if self.running:
@@ -355,7 +393,8 @@ class Macro:
         self._running_task = asyncio.create_task(self._do_run())
 
         try:
-            await self._running_task
+            async with OverheadHelper(self, "", macro_id=self.macro_id):
+                await self._running_task
         except asyncio.CancelledError:
             with suppress(asyncio.CancelledError):
                 self._running_task.cancel()
@@ -368,7 +407,6 @@ class Macro:
         return StageStatus.FAILED not in self.stage_status.values()
 
     def _get_coros(self, stage: StageType) -> list[Coroutine]:
-
         if isinstance(stage, str):
             stage_method = getattr(self, stage)
 
@@ -390,14 +428,39 @@ class Macro:
         current_task: asyncio.Future | None = None
 
         for istage, stage in enumerate(self.stages):
-            stage_coros = [asyncio.create_task(coro) for coro in self._get_coros(stage)]
-            current_task = asyncio.gather(*stage_coros)
+            coros = self._get_coros(stage)
+            wrapped_coros = [
+                asyncio.create_task(record_overhead(self)(coro)) for coro in coros
+            ]
+
+            # If we are running multiple stages concurrently, we also record the
+            # overhead of the entire set.
+            overhead_helper: OverheadHelper | None = None
+            if len(wrapped_coros) > 1:
+                costage_name = ":".join([coro.__name__ for coro in coros])
+                overhead_helper = OverheadHelper(
+                    self,
+                    costage_name,
+                    macro_id=self.macro_id,
+                )
+                await overhead_helper.start()
+
+            current_task = asyncio.gather(*wrapped_coros)
 
             self.set_stage_status(stage, StageStatus.ACTIVE)
 
             try:
                 await current_task
 
+                was_cancelling = self.has_status(stage, StageStatus.CANCELLING)
+
+                # Regardless of whether it was cancelling, this stage finished.
+                self.set_stage_status(stage, StageStatus.FINISHED)
+
+                # But now abort the macro.
+                if was_cancelling:
+                    raise asyncio.CancelledError()
+
             except asyncio.CancelledError:
                 with suppress(asyncio.CancelledError):
                     current_task.cancel()
@@ -416,7 +479,10 @@ class Macro:
                     output=False,
                 )
                 self.cancelled = True
-                await self.fail_macro(MacroError("The macro was cancelled"))
+
+                # Not really failing the macro since this was a user
+                # requested cancellation.
+                await self.fail_macro("The macro was cancelled")
                 return
 
             except Exception as err:
@@ -427,22 +493,50 @@ class Macro:
 
                 # Cancel stage tasks (in case we are running multiple concurrently).
                 with suppress(asyncio.CancelledError):
-                    for task in stage_coros:
+                    for task in wrapped_coros:
                         if not task.done():
                             task.cancel()
 
                 await self.fail_macro(err, stage=stage)
                 return
 
-            self.set_stage_status(stage, StageStatus.FINISHED)
+            finally:
+                if overhead_helper is not None:
+                    await overhead_helper.stop()
 
-    def cancel(self):
+    def get_active_stages(self):
+        """Returns a list of running stages."""
+
+        return [
+            stage
+            for stage in self.stage_status
+            if self.stage_status[stage] == StageStatus.ACTIVE
+        ]
+
+    def is_stage_done(self, stage: str):
+        """Returns `True` if a stage is finished."""
+
+        return self.has_status(stage, StageStatus.FINISHED)
+
+    def has_status(self, stages: StageType, status: StageStatus):
+        """Determines if any of the stages has that status."""
+
+        if isinstance(stages, str):
+            return self.stage_status[stages] == status
+
+        return any([self.stage_status[stage] == status for stage in stages])
+
+    def cancel(self, now: bool = True):
         """Cancels the execution ot the macro."""
 
         if not self.running or not self._running_task:
             raise MacroError("The macro is not running.")
 
-        self._running_task.cancel()
+        if now:
+            self._running_task.cancel()
+        else:
+            active_stages = self.get_active_stages()
+            self.set_stage_status(active_stages, StageStatus.CANCELLING)
 
     async def send_command(
         self,
@@ -478,3 +572,10 @@ class Macro:
                 raise MacroError(f"Command {target} {command_string} failed.")
 
         return command
+
+    async def wait_until_complete(self):
+        """Asynchronously blocks until the macros is done, cancelled, or failed."""
+
+        await self._running_event.wait()
+
+        return not self.failed
